package cs101.awt.geom;
/*
 * ShapeUtils.java
 *
 * Created on March 2, 2003, 2:30 PM
 */
import java.awt.Shape;
import java.awt.Point;
import java.awt.Rectangle;

import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import java.awt.geom.AffineTransform;
import java.awt.geom.IllegalPathStateException;
import java.awt.geom.NoninvertibleTransformException;

import java.util.ArrayList;
import java.util.List;


/**
 * This utility class holds routines for doing conversions on classes that
 * implement the <code>Shape</code> interface. Generally the first step is
 * in each routine is to convert the shape to a <code>GeneralPath</code>, and
 * return values of type <code>Shape</code> will in fact be instances of
 * the <code>GeneralPath</code> class unless specified otherwise.
 *
 * <a id="knotpts"></a>
 * <p>Java's <code>GeneralPath</code> class is a series of concatenated line
 * segments, second order (quadratic) Bezier curves and third order (cubic)
 * Bezier curves. Bezier Curves consist of two types of points: <b>knot
 * points</b> which are points that lie on the curve, and <b>control points</b>
 * which do not lie on the curve but denote vectors that influence the shape
 * of the curve between knot points.
 *
 * <p>None of the routines in this class attempt to convert between lines,
 * quadratics, and cubics. A shape consisting entirely of lines, will remain
 * entirely line based, and a purely cubic shape will also remain cubic. Mixed
 * shapes will remain mixed, but the ratios of the number of segements of any
 * two types is almost certain to change.
 *
 * @author  Patrick G. Heck, Gus.heck@olin.edu
 * @version $Id: ShapeUtils.java,v 1.18 2003/09/23 15:23:40 gus Exp $
 */
public final class ShapeUtils {
    
    /** Creates a new instance of ShapeUtils */
    private ShapeUtils() {
    }
    
    /**
     * Compute a velocity vector after a bounce off a given surface. The bounce
     * is computed by rotating the system until the line is vertical,
     * inverting the x component to calculate the bounce, and then performing
     * the inverse transform to get the desired.
     *
     * @param surface The line representing the angle of the surface to bounce
     *                off of.
     * @param velVect A line with starting point 0,0 representing the velocity
     *                vector to be bounced.
     * @return  A line starting at 0,0 representing the new velocity vector.
     */
    public static Line2D.Float bounce(Line2D.Float surface, Line2D.Float velVect) {
        AffineTransform rotateVert, rotateBack;
        double lineSlant = angleToVert(surface);

        rotateVert = AffineTransform.getRotateInstance(lineSlant);
        
        try {
            rotateBack = rotateVert.createInverse();
        } catch (NoninvertibleTransformException nte) {
            System.out.println("NoninvertableTransformException! bounce not calculated.");
            throw new Error("Rotations shoud be invertable!");
        }
        
        // this should rotate the line around 0,0 which is the x1,y1 point
        Shape s = rotateVert.createTransformedShape(velVect);
        PointIterator iter = new PointIterator(s);
        velVect = new Line2D.Float(iter.nextPoint(),iter.nextPoint());
        
        // bouncing off a vertical line is just a matter of inverting the
        // x component
        velVect.setLine(0,0,-velVect.getX2(),velVect.getY2());
        
        s = rotateBack.createTransformedShape(velVect);
        iter = new PointIterator(s);
        velVect = new Line2D.Float(iter.nextPoint(),iter.nextPoint());
        
        return velVect;
    }
    
    
    /**
     * Find the resultant velocity from the collision of a mobile and imobile
     * shape. The two shapes must have already collided and be in a state of
     * overlap, and must be shapes consisting of more than one point.
     *
     * @param s1 The moving shape (tiny mass)
     * @param s2 The imobile shape (infinite mass)
     * @param v1X The x component of the velocity of s1
     * @param v1Y The y component of the velocity of s1
     * @return a line starting at the origin and representing the new velocity
     *         vector
     */
    public static Line2D.Float doReflect(Shape s1, Shape s2, double v1X, double v1Y) {
        if (!isOverlapping(s1,s2)) {
            //System.exit(1);
            throw new ShapesDontOverlapException("Cannot reflect shapes until"+
            "they hit each other");
        }
        Line2D.Float nearSeg = null;
        
        Line2D.Float velVect = new Line2D.Float(0,0,
        new Double(v1X).floatValue(),
        new Double(v1Y).floatValue());
        Point2D.Float insidePt;
        try {
            insidePt = meanContainedPoint(s2,s1); // points conatained by stationary object
        } catch (NotEnoughPointsException nepe) {
            try {
            insidePt = meanContainedPoint(s1,s2); 
            } catch (NotEnoughPointsException nepe2) {
                throw new Error("overlapping but neither conatins the other point!");
            }
        }
        try {
            nearSeg = nearestSegment(s2,insidePt);
        } catch (NotEnoughPointsException nepe) {
            System.out.println("Ignoring malformed shape:" + nepe.getMessage());
            return velVect; // pretend malformed shapes arn't there
        } catch (NoUniqueLineException nule) {
            Point2D.Float nearest = getNearestPoint(s1,insidePt);
            nearSeg = getLineFromNeighbors(s1,nearest);
        }
        
        return bounce(nearSeg, velVect);
        
    }
    
    
    /**
     * Find the resultant velocity from the collision of a mobile and imobile
     * shape. The two shapes must have already collided and be in a state of
     * overlap, and must be shapes consisting of more than one point.
     *
     * @param s1 The moving shape (tiny mass)
     * @param s2 The imobile shape (infinite mass)
     * @param velVect1 The velocity vector of s1, (x1,y1 point at 0.0 x2,y2
     *                 representing the x and y components of the velocity)
     * @return a line starting at the origin and representing the new velocity
     *         vector
     */
    public static Line2D.Float doReflect(Shape s1, Shape s2, Line2D.Float velVect1) {
        return doReflect(s1,s2,velVect1.getX2(),velVect1.getY2());
    }
    
    /**
     * Find the new x velocity after reflecting a moving shape from a imobile
     * shape.
     *
     * @param s1 The moving shape (tiny mass)
     * @param s2 The imobile shape (infinite mass)
     * @param v1X The x component of the velocity of s1
     * @param v1Y The y component of the velocity of s1
     */
    public static double doXReflect(Shape s1, Shape s2, double v1X, double v1Y) {
        
        return doReflect(s1,s2,v1X,v1Y).getX2();
        
    }
    
    /**
     * Find the new x velocity after reflecting a moving shape from a imobile
     * shape.
     *
     * @param s1 The moving shape (tiny mass)
     * @param s2 The imobile shape (infinite mass)
     * @param v1X The x component of the velocity of s1
     * @param v1Y The y component of the velocity of s1
     */
    public static double doYReflect(Shape s1, Shape s2, double v1X, double v1Y) {
        
        return doReflect(s1,s2,v1X,v1Y).getY2();
        
    }
    
    /**
     * Find the angle between a given line segment and the vertical axis in
     * direction that roataes the positive x axis towards positive y axis
     *
     * @param aLine  A line to find the angle of
     * @return The angle in radians.
     */
    public static double angleToVert(Line2D.Float aLine) {
        
        Point2D.Float start = new Point2D.Float();
        Point2D.Float end  = new Point2D.Float();
        Point2D.Float near, far;
        
        start.setLocation(aLine.getX1(),aLine.getY1());
        end.setLocation(aLine.getX2(),aLine.getY2());
        
        double rise = start.y - end.y;
        double run = start.x - end.x;
        
        double tanTheta = rise/run;
            
        return Math.PI/2 - Math.atan(tanTheta);
        
    }
    
    /**
     * Find the line between points on either side of a knot point on a given
     * shape. If the shape is unclosed, and the point queried is the start
     * or end of the shape then the line between the querried point and the
     * nearest point will be returned. The shape must have more than one point.
     *
     * @param s The shape to find points from, comprised of than one point.
     * @param pt The point that is a knot of s for which we want a neighbors
     *           line.
     */
    
    public static Line2D.Float getLineFromNeighbors(Shape s, Point2D.Float pt) {
        PointIterator iter = new PointIterator(s);
        Point2D.Float preceding = null;
        Point2D.Float following = null;
        Point2D.Float prevPt = null;
        Point2D.Float currPt = iter.nextPoint();
        Point2D.Float startPt = currPt;
        
        while (iter.hasNext() && iter.isStart(startPt)) {
            if (startPt.equals(pt)) {
                if (iter.hasNext()) {
                    following = iter.nextPoint();
                    currPt = following;
                } else {
                    throw new IllegalArgumentException("Shape has only one Point!");
                }
                while (iter.hasNext() && iter.isStart(startPt) && (currPt != startPt)) {
                    prevPt = currPt;
                    currPt = iter.nextPoint();
                }
                if (currPt == startPt) {       // closed shape
                    return new Line2D.Float(prevPt,following);
                }
                
                // open or discontinuous shape
                return new Line2D.Float(startPt,following);
                
            }
            prevPt = currPt;
            currPt = iter.nextPoint();
            
            if (currPt.equals(pt)) {
                preceding = prevPt;
                if(iter.hasNext()) {
                    currPt = iter.nextPoint();
                }
                return new Line2D.Float(preceding,currPt);
            }
            
        }
        
        throw new IllegalArgumentException("Point was not a knot of the supplied Shape");
        
    }
    
    /**
     * Find the knot point in the given shape that is closest to the specified
     * point. In the case of equidistant points, the qualifying point closest
     * to the start of the shape will be returned. The shape passed to this
     * method must have at least one point.
     *
     * @param s The shape to find a point from
     * @param pt The point to which the points of the shape are compared.
     * @return A point on the shape that is as close or closer than any other
     *         point on the shape to pt
     */
    
    public static Point2D.Float getNearestPoint(Shape s, Point2D.Float pt) {
        Point2D.Float result = null;
        Point2D.Float test = null;
        PointIterator iter = new PointIterator(s);
        
        // if you pass an empty shape it is your problem not mine doofus :)
        result = iter.nextPoint();
        while(iter.hasNext()) {
            test = iter.nextPoint();
            if (pt.distance(result) > pt.distance(test)) {
                result = test;
            }
        }
        return result;
    }
    
    /**
     * Find the line segment between two consecutive knot points of the given
     * <code>Shape</code> that is closest to a specified point. Note that this
     * method will return a line even if the segment it represents in the shape
     * is a cubic or quadratic bezier curve.
     *
     * @param aShape The shape from which to find the line segment
     * @param aPoint  The point to which the distance is compared.
     * @return A unique line segment that is closer to the point than any other.
     *
     * @throws NoUniqueLineException If there are two or more lines equidistant
     *                               from the point.
     * @throws NotEnoughPointsException If the aShape contains less than 2 points
     *
     */
    public static Line2D.Float nearestSegment(Shape aShape,
    Point2D.Float aPoint) throws NoUniqueLineException,
    NotEnoughPointsException {
        
        PointIterator iter = new PointIterator(aShape);
        Point2D.Float last = null;
        Point2D.Float curr = null;
        Line2D.Float closest = null;
        Line2D.Float nextBest = null;
        Line2D.Float other = null;
        double closestDist = -1;
        double nextBestDist = -1;
        
        if (iter.hasNext()) {
            last = iter.nextPoint();
        }
        if (iter.hasNext()) {
            curr = iter.nextPoint();
        }
        
        if (last == null || curr == null) {
            throw new NotEnoughPointsException("Not enough points for a line");
        }
        
        if (last.equals(curr) && !iter.hasNext()) {
            throw new NotEnoughPointsException("Only two duplicate points");
        }
        
        closest = new Line2D.Float(last,curr);
        closestDist = closest.ptSegDist(aPoint);
        while (iter.hasNext()) {
            last = curr;
            curr = iter.nextPoint();
            if (last.equals(curr)) {
                continue;
            }
            other = new Line2D.Float(last,curr);
            double otherDist = other.ptSegDist(aPoint);

            if (nextBest == null) {
                nextBest = other;
                nextBestDist = otherDist;
            }
            
            if (closestDist == otherDist) {
                nextBest = other;
                nextBestDist = otherDist;
            } else {
                if (closestDist > otherDist) {
                    nextBest = closest;
                    nextBestDist = closestDist;
                    closest = other;
                    closestDist = otherDist;
                }
            }
        }
        
        // The following case would occr with a shape constructed with
        // moveTo(1,1) lineTo(1,1) closePath() or other "go nowhere" shapes.
        if (nextBest == null) {
            throw new NotEnoughPointsException("Only duplicate points found");
        }
        
        if (closestDist == nextBestDist) {
            throw new NoUniqueLineException("More than one Line");
        }
        
        return closest;
        
    }
    
    /**
     * Find the mean location of a list of <code>Point2D.Float</code>.
     *
     * @param pointList A list containing <em>only</em> Point2D.Float
     * @return The average point.
     * @throws ClassCastException If the objects in the supplied list are not
     *                            of type Point2D.Float
     */
    public static Point2D.Float meanPoint(List pointList) {
        double sumX = 0;
        double sumY = 0;
        float meanX, meanY;
        
        for (int i=0; i<pointList.size();i++) {
            sumX += ((Point2D.Float) pointList.get(i)).x;
            sumY += ((Point2D.Float) pointList.get(i)).y;
        }
        
        meanX = new Double(sumX/pointList.size()).floatValue();
        meanY = new Double(sumY/pointList.size()).floatValue();
        
        return new Point2D.Float(meanX, meanY);
        
    }
    
    
    /**
     * Calcualte the average of all points contained within one shape and
     * belonging to another.
     *
     * @param container The shape which may contain points
     * @param containee The shape whose points may be contained
     * @return          The average of the points contained.
     */
    public static Point2D.Float meanContainedPoint(Shape container, Shape containee) throws NotEnoughPointsException {
        ArrayList contained = getContainedPoints(container, containee);
        if (contained.size() == 0) {
            throw new NotEnoughPointsException("no points contained by container");
        }
        return meanPoint(contained);
    }
    
    /**
     * Get a list of the points belonging to one shape and contained within
     * another.
     *
     * @param container The shape which may contain points
     * @param containee The shape whose points may be contained
     * @return          The points belonging to containee and within conainer
     */
    public static ArrayList getContainedPoints(Shape container, Shape containee) {
        ArrayList points = new ArrayList();
        PointIterator iter = new PointIterator(containee);
        Point2D.Float temp;
        
        while (iter.hasNext()) {
            temp = iter.nextPoint();
            if (container.contains(temp.x,temp.y)) {
                points.add(temp.clone());
            }
        }
        
        return points;
    }
    
    /**
     * Test to see if two <code>Shape</code> objects overlap.
     * The algorithim in this test will only detect the type of
     * overlap where a <a href="knotpts">knot point</a> from one shape is
     * contained by the other shape (as determined by each shape's
     * implementation of the <code>contains(double, double)</code> method.
     *
     * @param shapeA the first shape
     * @param shapeB the second shape
     * @return true if the shapes overlap false otherwise
     */
    
    public static boolean isOverlapping(Shape shapeA, Shape shapeB) {
        PointIterator iter = new PointIterator(shapeA);
        PointIterator iter2 = new PointIterator(shapeB);
        
        Point2D.Float temp;
        
        // Begin CPU time optimization:
        // ----------------------------
        // Most comparisons are between distant objects, so itterating
        // all points is unnecessary if the bounding boxes don't overlap. 
        
        Rectangle rectA = shapeA.getBounds();
        Rectangle rectB = shapeB.getBounds();
        
        if(!rectA.intersects(rectB)) { 
            return false; 
        }
                
        // end CPU time optimization        
        
        while(iter.hasNext()) {
            temp = iter.nextPoint();
            if (shapeB.contains(temp.x,temp.y)) {
                return true;
            }
        }
        while(iter2.hasNext()) {
            temp = iter2.nextPoint();
            if (shapeA.contains(temp.x,temp.y)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Convert any AWT shape into a shape with a specified precision.
     * The specified precision indicates the maximum tolerable distance
     * between <a href="knotpts">knot points</a>. If the shape is already
     * precise enough then it is returned unmodified.
     *
     * @param aShape the shape to be converted if necessary.
     * @param precision the maximum tolerable distance between knot points.
     * @return A more precise version of the shape, or the same shape if
     *         the precision is already satisfied.
     */
    public static Shape getPreciseShape(Shape aShape, float precision) {
        GeneralPath path = new GeneralPath(aShape);
        GeneralPath newPath = getPrecisePath(path, precision);
        
        return (path == newPath) ? aShape : newPath;
    }
    
    
    /**
     * Increase the precision of a given <code>GeneralPath</code> object.
     * The specified precision indicates the maximum tolerable distance
     * between <a href="knotpts">knot points</a>. If the path is already
     * precise enough then it is returned unmodified.
     *
     * @param aPath The path to convert if necessary.
     * @param precision the maximum tolerable distance between knot points.
     * @return A more precise version of the path or the same path if the
     *         precision was alredy satisfied.
     */
    public static GeneralPath getPrecisePath(GeneralPath aPath, float precision) {
        GeneralPath[] paths = new GeneralPath[2];
        
        paths[0] = aPath;
        while (paths[0] != paths[1]) {
            paths[1] = paths[0];
            paths[0] = getImprovedPath(paths[0],precision);
        }
        return paths[0];
    }
    
    
    /**
     * Used iteratively by getPrecisePath to acheive as specified precision.
     * The path is traversed and for every case in which the distance between
     * consecutive <a href="knotpts">knot points</a> is greater than the
     * precision value, a new knot point and associated control points are
     * inserted. The surroundng knot points have their control points modified
     * appropriately to maintain an Identical curve. The new knot point is
     * only guaranteed to lie on the curve and in extreme cases may even be
     * further from either of the orriginal knot points than the distance
     * between the original knot points, but this indicates a situation in
     * which a radical curve has been drawn and further itterations
     * will eventually reduce the distance.<p>
     *
     * Note that because GeneralPath does not overide equals we
     * return a reference to the SAME object passed in if there are no
     * improvements to be made. aPath.equals(newPath) is equivalent to
     * aPath == newPath.
     *
     * @param aPath      The path that may need to be improved.
     * @param precision  The minimum acceptable distance between knot points.
     */
    public static GeneralPath getImprovedPath(GeneralPath aPath, float precision) {
        GeneralPath newPath = new GeneralPath();
        PathIterator iter = aPath.getPathIterator(null);
        Point2D.Float currPt = new Point2D.Float();
        Point2D.Float conPt1 = new Point2D.Float();
        Point2D.Float conPt2 = new Point2D.Float();
        Point2D.Float nextPt = new Point2D.Float();
        Point2D.Float pathStart = new Point2D.Float();
        //Point2D.Float floatPt = new Point2D.Float();
        float[] seg = new float[12];
        int segType;
        boolean improved = false;
        
        // flags to do checking that Sun should have done in GeneralPath.
        // As of 1.4.1 Sun only guarantees that the path starts with a
        // segment of type SEG_MOVETO.
        boolean closed = false;
        
        while(!iter.isDone()) {
            segType = iter.currentSegment(seg);
            iter.next();
            switch (segType) {
                case PathIterator.SEG_MOVETO :
                    closed = false;
                    pathStart.setLocation(seg[0], seg[1]);
                    currPt.setLocation(seg[0], seg[1]);
                    nextPt.setLocation(seg[0], seg[1]);
                    newPath.moveTo(nextPt.x,nextPt.y);
                    break;
                case PathIterator.SEG_CLOSE :
                    if (closed) {
                        throw new IllegalPathStateException("Path closed twice!");
                    }
                    nextPt.x = pathStart.x;
                    nextPt.y = pathStart.y;
                    closed = true;
                    
                    if (currPt.distance(nextPt) <= precision) {
                        newPath.closePath();
                    } else {
                        improved = true;
                        seg[0] = nextPt.x;
                        seg[1] = nextPt.y;
                        
                        createIntermediateLine(currPt,seg); // seg modified
                        newPath.lineTo(seg[0],seg[1]);
                        newPath.closePath();
                    }
                    break;
                    
                case PathIterator.SEG_LINETO :
                    if (closed) {
                        throw new IllegalPathStateException("Path already closed!");
                    }
                    nextPt.setLocation(seg[0],seg[1]);
                    
                    if (currPt.distance(nextPt) <= precision) {
                        newPath.lineTo(nextPt.x,nextPt.y);
                    } else {
                        improved = true;
                        
                        createIntermediateLine(currPt,seg); // seg modified
                        newPath.lineTo(seg[0],seg[1]);
                        newPath.lineTo(seg[2],seg[3]);
                    }
                    break;
                case PathIterator.SEG_QUADTO :
                    if (closed) {
                        throw new IllegalPathStateException("Path already closed!");
                    }
                    conPt1.setLocation(seg[0],seg[1]);
                    nextPt.setLocation(seg[2],seg[3]);
                    
                    if (currPt.distance(nextPt) <= precision) {
                        newPath.quadTo(conPt1.x,conPt1.y,nextPt.x,nextPt.y);
                    } else {
                        improved = true;
                        
                        createIntermediateQuad(currPt,seg); // seg modified
                        newPath.quadTo(seg[0],seg[1],seg[2],seg[3]);
                        newPath.quadTo(seg[4],seg[5],seg[6],seg[7]);
                    }
                    break;
                case PathIterator.SEG_CUBICTO :
                    if (closed) {
                        throw new IllegalPathStateException("Path already closed!");
                    }
                    conPt1.setLocation(seg[0], seg[1]);
                    conPt2.setLocation(seg[2], seg[3]);
                    nextPt.setLocation(seg[4], seg[5]);
                    
                    if (currPt.distance(nextPt) <= precision) {
                        newPath.curveTo(conPt1.x,conPt1.y,conPt2.x,conPt2.y,nextPt.x,nextPt.y);
                    } else {
                        improved = true;
                        
                        createIntermediateCurve(currPt,seg); // seg modified
                        newPath.curveTo(seg[0],seg[1],seg[2],seg[3],seg[4],seg[5]);
                        newPath.curveTo(seg[6],seg[7],seg[8],seg[9],seg[10],seg[11]);
                        // add an intermediate point
                    }
                    break;
                default:
                    // should never happen unless Sun changes GeneralPath
                    throw new Error("Unknown segment type from Path Iterator");
            }
            
            currPt.setLocation(nextPt.x,nextPt.y);
        }
        
        
        if (!improved) {
            return aPath;
        } else {
            return newPath;
        }
    }
    
    /**
     * Mutate the suppled array of float to contain an additional knot.
     *
     * @param start the start point of the Line.
     * @param pts An array of at least size 4 containing the end point x,y
     *            coordinates at position 0 and 1 respectively. This array
     *            format matches that returned by <code>PathIterator</code>
     */
    public static void createIntermediateLine(Point2D.Float start, float[] pts) {
        Point2D.Float newPt = midPoint(start.x,start.y,pts[0],pts[1]);
        pts[2] = pts[0];
        pts[3] = pts[1];
        pts[0] = newPt.x;
        pts[1] = newPt.y;
    }
    
    
    /**
     * Mutate the suppled array of float to contain an additional knot and
     * modify the associated control points.
     *
     * @param start the start point of the quadratic Bezier spline.
     * @param pts An array of at least size 8 containing the end point x,y
     *            coordinates at position 2 and 3 respectively. The control
     *            point should be specified at positions 0 and 1. This array
     *            format matches the format returned by
     *            <code>PathIterator</code>.
     */
    public static void createIntermediateQuad(Point2D.Float start, float[] pts) {
        Point2D.Float nearCtl = midPoint(start.x,start.y,pts[0],pts[1]);
        Point2D.Float farCtl = midPoint(pts[0],pts[1],pts[2],pts[3]);
        
        Point2D.Float knot = midPoint(nearCtl.x,nearCtl.y,farCtl.x,farCtl.y);
        
        // move the next point to make space for the new data
        pts[6] = pts[2];
        pts[7] = pts[3];
        
        // record in the new points
        pts[0] = nearCtl.x;
        pts[1] = nearCtl.y;
        pts[2] = knot.x;
        pts[3] = knot.y;
        pts[4] = farCtl.x;
        pts[5] = farCtl.y;
    }
    
    /**
     * Mutate the supplied array of float to contain an additional knot and
     * modify the associated control points.
     *
     * @param start the start point of the quadratic Bezier spline.
     * @param pts An array of at least size 8 containing the end point x,y
     *            coordinates at position 4 and 5 respectively. The control
     *            point for the start point should be specified at positions
     *            0 and 1; The control point for the end point should be
     *            specified at positions 2 and 3 This array format matches the
     *            format returned by <code>PathIterator</code>.
     */
    public static void createIntermediateCurve(Point2D.Float start, float[] pts) {
        Point2D.Float stCtl = midPoint(start.x,start.y,pts[0],pts[1]);
        Point2D.Float endCtl = midPoint(pts[2],pts[3],pts[4],pts[5]);
        Point2D.Float btwCtls = midPoint(pts[0],pts[1],pts[2],pts[3]);
        
        Point2D.Float nearCtl = midPoint(stCtl.x,stCtl.y,btwCtls.x,btwCtls.y);
        Point2D.Float farCtl = midPoint(btwCtls.x,btwCtls.y,endCtl.x,endCtl.y);
        
        Point2D.Float knot = midPoint(nearCtl.x,nearCtl.y,farCtl.x,farCtl.y);
        
        // move the next point to make space for the new data
        pts[10] = pts[4];
        pts[11] = pts[5];
        
        // record in the new points
        pts[0] = stCtl.x;    // the modified start control
        pts[1] = stCtl.y;
        pts[2] = nearCtl.x;  // the start side control for the new knot
        pts[3] = nearCtl.y;
        pts[4] = knot.x;     // the new knot
        pts[5] = knot.y;
        pts[6] = farCtl.x;   // the end side control for the new knot
        pts[7] = farCtl.y;
        pts[8] = endCtl.x;   // the modified end control
        pts[9] = endCtl.y;
    }
    
    
    /**
     * Find the midpoint of a specifed line segment.
     *
     * @param x1 start x coordinate
     * @param y1 start y coordinate
     * @param x2 end x coordinate
     * @param y2 end y coordinate
     * @return The calculated midpoint
     */
    public static Point2D.Float midPoint(float x1, float y1, float x2, float y2) {
        return new Point2D.Float((x1+x2)/2, (y1+y2)/2);
    }
}

/*
 * $Log: ShapeUtils.java,v $
 * Revision 1.18  2003/09/23 15:23:40  gus
 * javadoc fixes
 *
 * Revision 1.17  2003/06/30 23:12:21  gus
 * remove a slew of nasty debugging code that was all
 * commented out anyway. At this point, I certainly
 * don't remember what most of it did anyway.
 *
 * Revision 1.16  2003/06/30 23:06:53  gus
 * Remove old broken code comment.
 *
 * Revision 1.15  2003/06/27 15:32:58  krivard
 * Bugfix:
 * Small boxes approaching large boxes from the left were falsely declared not overlapping.
 * Changed upperleft and lowerleft bounding box check in isOverlapping(Shape,Shape) to Rectangle.intersects().
 *
 * Revision 1.14  2003/03/10 16:47:24  gus
 * Added an optimization to isOverlapping so that points are not itterated unless the
 * bounding box of the two shapes overlapps. This creates a VERY noticiable
 * improvement in the breakout problem set.
 *
 * Revision 1.13  2003/03/10 15:53:56  gus
 * Commented out debugging prints
 * meanContainedPoint now  throws NotEnoughPointsException
 * Fixed nearestSegment
 * angleToVert now returns PI/2-arcTan(tanTheta)
 *
 * Revision 1.12  2003/03/07 19:20:37  gus
 * Fix npe that occurs when start point is the point around which we are trying to find
 * a line.
 *
 * Revision 1.11  2003/03/06 23:28:30  gus
 * Some reflection code that might even be useful
 *
 * Revision 1.10  2003/03/06 21:19:25  gus
 * A get nearest point method
 *
 * Revision 1.9  2003/03/06 20:52:44  gus
 * get rid of a class cast exception. CreateTransformedShape always returns
 * a GeneralPath, so cant' cast to Line2D.Float even if we know it is a line.
 *
 * Revision 1.8  2003/03/06 19:54:09  gus
 * make methods static cause I forgot to
 *
 * Revision 1.7  2003/03/06 19:07:07  gus
 * meant != rather than ==
 * fixed now.
 *
 * Revision 1.6  2003/03/06 18:55:21  gus
 * We need to advance the Pathe iterator despite what the nutshell book claims
 *
 * Revision 1.5  2003/03/06 18:24:39  gus
 * some partially complete bouncing routines
 *
 * Revision 1.4  2003/03/04 23:20:42  gus
 * bunch more methods added
 *
 * Revision 1.3  2003/03/04 16:04:45  gus
 * Javadoc improvements
 *
 * Revision 1.2  2003/03/04 15:32:20  gus
 * adding log comment
 *
 */
