/*
 * Coord.java
 * part of the Spirograph.java problem set
 *
 * Developed for "Rethinking CS101", a project of Lynn Andrea Stein's AP Group.
 * For more information, see <a href="http://www.ai.mit.edu/projects/cs101/">the
 * CS101 homepage</a> or email <las@ai.mit.edu>.
 *
 * Copyright (C) 1996 Massachusetts Institute of Technology.
 * Please do not redistribute without obtaining permission.
 */

package spirograph;

import java.util.*;
import java.awt.*;

/** This class holds information regarding one dimension of the ball.
 * It has methods to update the position of the ball, as well as
 * methods to get and set certain variables.
 *
 * <p>Copyright &copy; 1996 Massachusetts Institute of Technology.<br />
 * Copyright &copy; 2003 Franklin W. Olin College of Engineering.</p>
 *
 * @author Luis Sarmenta, lfgs@cag.lcs.mit.edu
 * @author Henry Wong, henryw@mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: Coord.java,v 1.4 2004/02/09 20:55:03 gus Exp $
 * @see AccelHandler
 * @see Spirograph
 */
public class Coord {
    
  private double pos = 0;
  private double vel = 0;
  private double accel = 0;
  private double maxPos = Spirograph.WIDTH;

  private int accelMode = AccelHandler.POSMODE;

  /** 
   * Determines the current mode of operation, either position,
   *       velocity, or accleration control.
   *
   * @return {@link AccelHandler#POSMODE}, {@link AccelHandler#VELMODE}
   *         or {@link AccelHandler#ACCELMODE}
   */
  public int getMode() { return accelMode; }

  /**
   * Sets the current mode.  Legal parameter values are
   *       AccelHandler.POSMODE, AccelHandler.VELMODE, or
   *       AccelHandler.ACCELMODE.
   *
   * @param mode The desired mode
   */
  public void setMode(int mode) {
    switch( mode ) {
    case AccelHandler.POSMODE: 
      setVel(0.0D);
      // fall through

    case AccelHandler.VELMODE:
      setAccel(0.0D);
      // fall through

    default:
      accelMode = mode;
      break;
    }
  }

  /** The maximum position for this coordinate.
   * @return a positive double indicating the maximum allowable position
   */  
  public double getMaxPos() { return maxPos; }
  /** Set the maximum position for this coordinate
   * @param pos A positive value indicating the maximum position.
   * @throws IllegalArgumentException when pos < 0
   */  
  public void setMaxPos( double pos ) throws IllegalArgumentException { 
      if (pos < 0.0) {
          throw new IllegalArgumentException("Negagive Max Pos!");
      }
      maxPos = pos; 
  }
	

  /**
   * The current position of this coordinate. 
   *
   * @return  a double between <code>-(getMaxPos())</code> and 
   *          <code>getMaxPos()</code>
   */  
  public double getPos() { return pos; }
  /** Modify the current position of this coordinate.
   *
   * @param pos the desired new position
   * @throws IllegalArgumentException If the absolute value of the argument is
   *                                  greater than {@link #getMaxPos()}
   */  
  public void setPos(double pos) throws IllegalArgumentException { 
      if (Math.abs(pos) > getMaxPos()) {
          throw new IllegalArgumentException("pos out of bounds");
      }
      this.pos = pos; 
  }

  /** The current velocity along this coordinate axis.
   *
   * @return the magnitude of a velocity vector along this axis
   */  
  public double getVel() { return vel; }
   /** Set the velocity along this axis.
    *
    * @param vel The new value for the velocity along this axis
    */  
  public void setVel(double vel) { this.vel = vel; }

  /** The acceleration along this axis.
   *
   * @return the magintude of an acceleration vector along this axis
   */  
  public double getAccel() { return accel; }
  /** Set the acceleration along this axis.
   *
   * @param accel The new value for the acceleration along this axis
   */  
  public void setAccel(double accel) { this.accel = accel; }                

  /** Sets the position, velocity or acceleration depending on the current mode.
   *
   * @param val The new value for the current movement mode.
   */  
  public void setValue(double val)
  {
    switch( accelMode ) {
    case AccelHandler.POSMODE: 
      setPos(val);
      break;
    case AccelHandler.VELMODE: 
      setVel(val);
      break;
    case AccelHandler.ACCELMODE: 
      setAccel(val);
      break;
    }
  }

  /** Calculates the acceleration of the ball by taking
   *  the current acceleration and adding it to the combined
   *  accelerations resulting from each of the different gravitational
   *  fields.
   *
   * @param v A vector denoting the distance on this axis and the total distance from
   *         every gravitational source.
   * @param step The size of the step to calculate.
   */
  public void nextStep (Vector v, double step) {
    double finalAccel = accel;
	
    for (Enumeration e = v.elements(); e.hasMoreElements();) {
      Point data  = (Point)e.nextElement();

      // I just use these to make the math formula a little
      // easier to follow, they aren't really needed.
      int dx = data.x;
      int dr = data.y;

      // If dr == 0 just ignore it...
      if (dr != 0) {	    
	finalAccel += Spirograph.GRAV*dx/Math.pow(dr,3);
      }

    }
	
    vel = vel + (finalAccel * step);
    pos = pos + (vel * step);
  }
}

/*
 * $Log: Coord.java,v $
 * Revision 1.4  2004/02/09 20:55:03  gus
 * javadoc fixes
 *
 * Revision 1.3  2003/01/15 17:36:10  gus
 * adding log keywords to files that don't have them
 *
 */