/*
 *  DotFrame.java
 *	Part of the Spirograph problem set.
 *
 * Developed for "Rethinking CS101", a project of Lynn Andrea Stein's AP Group.
 * For more information, see <a href="http://www.ai.mit.edu/projects/cs101/">the
 * CS101 homepage</a> or email <las@ai.mit.edu>.
 *
 * Copyright (C) 1998 Massachusetts Institute of Technology.
 * Please do not redistribute without obtaining permission.
 */

package spirograph;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/** This class pops up the window for the Spirograph, creates all of
 * the buttons and the DotPanel, and adds them. It also creates an
 * instance of AdvEnv and passes it a listener that will set the
 * velocity of the dot. It also starts and suspends the three
 *  Threads (one for each of the AccelHandlers and one for the
 *  DotPanel).</p>
 *
 * <p>Copyright &copy; 1998 Massachusetts Institute of Technology.<br />
 * Copyright &copy; 2002-2003 Franklin W. Olin College of Engineering.</p>
 *
 * @author Luis Sarmenta, lfgs@cag.lcs.mit.edu
 * @author Henry Wong, henryw@mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: DotFrame.java,v 1.7 2004/02/09 20:55:03 gus Exp $
 * @see Coord
 * @see DotPanel
 * @see AccelHandler
 * @see AdvEnv
 * @see SetVListener
 * @see Spirograph
 */
public class DotFrame extends Frame {

  // Must be fields for anonymouse classes
  private Thread xSpirit, ySpirit, painter;
  private boolean xStarted = false, yStarted = false, pStarted = false;
  private DotPanel myPanel;
  private ResetMenu myRM;
  private Coord x, y;

  // These fields are used to calculate the difference in size
  // between the Panel and the Frame. They are used in order to resize
  // the Panel after the Frame has been resized.
    
  private int xDif;
  private int yDif;

  private AdvEnv myAdvEnv;
    
  /** Creates a new Dot Frame with the Default font.
   *
   * @param xHandler The {@link AccelHandler} for the x direction
   * @param yHandler The {@link AccelHandler} for the y direction
   * @param xCoord The {@link Coord} object that defines our movement vectors on the x axis
   * @param yCoord The {@link Coord} object that defines our movement vectors on the y axis
   */
  public DotFrame(AccelHandler xHandler, AccelHandler yHandler, Coord xCoord, Coord yCoord) {
    this(xHandler, yHandler, xCoord, yCoord,
	 Spirograph.DEFAULTFONT);
  }
  
  /** Creates a new dot frame with the default font and <code>null</code> handlers.
   *
   * @param xCoord The {@link Coord} object that defines our movement vectors on the x axis
   * @param yCoord The {@link Coord} object that defines our movement vectors on the y axis
   */  
  public DotFrame(Coord xCoord, Coord yCoord) {
      this((AccelHandler)null, (AccelHandler)null, xCoord, yCoord, Spirograph.DEFAULTFONT);
  }
	
  /** Same as other constructor, except it sets the font rather than
   * using a default.
   *
   * @param xHandler The {@link AccelHandler} for the x direction
   * @param yHandler The {@link AccelHandler} for the y direction
   * @param xCoord The {@link Coord} object that defines our movement vectors on the x axis
   * @param yCoord The {@link Coord} object that defines our movement vectors on the y axis
   * @param f The desired font for this component
   */
  public DotFrame(AccelHandler xHandler, AccelHandler yHandler, Coord xCoord, Coord yCoord, Font f) {
	
    this.setFont(f);
    myPanel = new DotPanel(xCoord, yCoord);
	
    x = xCoord;
    y = yCoord;
		
    this.setLayout( new BorderLayout(10, 10)); 
    
    // Create the textfields and listeners for AdvEnv and pass them in.
    TextField xVel = new TextField(3);
    TextField yVel = new TextField(3);
    SetPosGravListener panelListener = new SetPosGravListener(x,y,myPanel);
    myPanel.addMouseListener(panelListener);
    
    myAdvEnv = new AdvEnv(xVel, yVel, myPanel);
    
    myRM = new ResetMenu();
    myRM.addListeners(new ActionListener() {
        public void actionPerformed(ActionEvent ae) {
            myPanel.flushLines();
            myPanel.paintBuf();
        }
    },
    new ActionListener() {
        public void actionPerformed(ActionEvent ae) {
            myPanel.flushGrav();
            myPanel.paintBuf();
        }
    },
    new ActionListener() {
        public void actionPerformed(ActionEvent ae) {
            suspendSpirits();
            suspendPainting();
            x.setPos(0);
            x.setVel(0);
            y.setPos(0);
            y.setVel(0);
            myPanel.paintBuf();
        }
    },
    new ActionListener() {
        public void actionPerformed(ActionEvent ae) {
            suspendSpirits();
            suspendPainting();
            x.setPos(0);
            x.setVel(0);
            y.setPos(0);
            y.setVel(0);
            myPanel.flushLines();
            myPanel.flushGrav();
            myPanel.paintBuf();
        }
    });
    
    if (xHandler == null) {
        this.setTitle("Etch-a-Sketch Drawing");
    }
    
    if (xHandler != null && yHandler != null) {
        
        this.setTitle("Spirograph");
        
        // Create the buttons and add some functionality.
        Button start=new Button("Start");
        start.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if(xStarted) {
                    xSpirit.resume();
                } else {
                    xSpirit.start();
                    xStarted = true;
                }
                if(yStarted) {
                    ySpirit.resume();
                } else {
                    ySpirit.start();
                    yStarted = true;
                }
                beginPainting();
            }
        });
        
        Button stop = new Button("Stop");
        stop.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ae) {
                xSpirit.suspend();
                ySpirit.suspend();
                painter.suspend();
            }
        });
        
        // Create the ResetMenu and some listeners for its various
        // buttons
        
        Button reset=new Button("Reset");
        reset.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                myRM.setVisible(true);
            }
        });
        
        
        
        
        // Add Listeners
        SetVListener myVelSetter = new SetVListener(xVel, yVel, x, y);
        myAdvEnv.addListener(myVelSetter);
        
        Button advEnv = new Button("Advanced Environment Options");
        advEnv.addActionListener(new ActionListener() {
            
            public void actionPerformed(ActionEvent ae) {
                myAdvEnv.setVisible(true);
            }
        });
        
        // Create the Quit Button
        Button quit=new Button("Quit");
        quit.addActionListener (new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                xSpirit.stop();
                ySpirit.stop();
                setVisible(false);
                dispose();
                System.exit(0);
            }
        });
        
        System.out.println ("Buttons created.");
        
        // Create the Button panel
        Panel p = new Panel();
        p.setLayout(new FlowLayout(FlowLayout.CENTER));
        
        p.add(start); 
        p.add(stop);
        p.add(reset);
        p.add(advEnv);
        p.add(quit);
        this.add("South",p);
        
        System.out.println ("Button panel created");
        
        xSpirit = new Thread (xHandler);
        ySpirit = new Thread (yHandler);
        
    }
    
    painter = new Thread (myPanel);

    this.add("Center", myPanel);
	
    // The Spirograph handles Resizing the window by changing the
    // size of the Panel. The variables xDif and yDif represents
    // the difference in size between the Frame and the Panel.
	    
    this.addComponentListener(new ComponentListener() {
      // if the window is resized, notify the appropriate
      // Objects.
		
      public void componentResized(ComponentEvent e) {
	int width = getSize().width-xDif;
	int height = getSize().height-yDif;
		    
	System.out.println ("Resized to " + width
			    + " wide and " +
			    height + " high.");
		
	myPanel.setSize(width,height);
      }

      // I need to implement these methods for the
      // ComponentListener interface.
      public void componentHidden(ComponentEvent e){}
      public void componentShown(ComponentEvent e){}
      public void componentMoved(ComponentEvent e){}
	    
    });
	    
    System.out.println ("Stuff added.");
	    
    this.pack();	
    this.setVisible(true);

    // xDif and yDif have to be set after Frame is made Visible,
    // otherwise the calls to getSize will return 0
	
    xDif = this.getSize().width - myPanel.getSize().width;
    System.out.println ("Frame width: " + xDif);
    yDif = this.getSize().height - myPanel.getSize().height;
    System.out.println ("Frame height: " + yDif);
	
    System.out.println ("Done.");
  }
  
  /** Show the advanced environment options. This method will be removed when
   * my AdvEtchControls is fully implemented. It is only used by the new
   * interface.
   *
   * @deprecated Since 1/14/03 - will be removed when AdvEtchControl is complete.
   */  
  public void showAdvEnv() {
      myAdvEnv.setVisible(true);
  }
  
  /** Show the advanced environment options. This method will be removed when
   * my AdvEtchControls is fully implemented. It is only used by the new
   * interface.
   *
   * @deprecated Since 1/14/03 - will be removed when AdvEtchControl is complete.
   */  
  public void showResetMenu() {
      myRM.setVisible(true);
  }
  
  /**
   * Suspend the the painting thread. This uses the dangerous and deprecated
   * <a href ="http://java.sun.com/j2se/1.4.1/docs/guide/misc/threadPrimitiveDeprecation.html">
   * Thread.resume()</a> method.The thread it can be restarted
   * with {@link DotFrame#beginPainting}
   *
   */  
  public void suspendPainting() {
      painter.suspend();
  }
  
  /**
   * Start or Resume painting actions. This uses the dangerous and deprecated
   * <a href ="http://java.sun.com/j2se/1.4.1/docs/guide/misc/threadPrimitiveDeprecation.html">
   * Thread.resume()</a> method. Painting Action can be
   * suspended with {@link DotFrame#suspendPainting()}
   *
   */  
  public void beginPainting() {
      if(pStarted) {
          painter.resume();
      } else {
          painter.start();
          pStarted = true;
      }
  }
  
  /** Suspend the the spirit threads. This uses the dangerous and deprecated
   * <a href ="http://java.sun.com/j2se/1.4.1/docs/guide/misc/threadPrimitiveDeprecation.html">
   * Thread.resume()</a> method.
   */  
  public void suspendSpirits() {
      if (xSpirit != null) {
          xSpirit.suspend();
      }
      if (ySpirit != null) {
          ySpirit.suspend();
      }
  }
  
  /** Stop the painting thread and call super.dispose()
   */  
  public void dispose() {
      painter.stop();
      super.dispose();
  }
}

/*
 * $Log: DotFrame.java,v $
 * Revision 1.7  2004/02/09 20:55:03  gus
 * javadoc fixes
 *
 * Revision 1.6  2003/01/15 17:36:10  gus
 * adding log keywords to files that don't have them
 *
 */